from django.forms.widgets import MediaDefiningClass
from django.template import Context, Template
from django.template.loader import get_template
from django.utils.safestring import mark_safe

from .utils import construct_template_name, flatstyle, flatatt

VALID_RENDER_FORMAT = ["html", "dict", "json"]


class NotComponent(Exception):
    pass


class InvalidComponentTemplate(Exception):
    pass


class BaseComponent:

    template_string = None

    def __init__(self, classes=None, attrs=None, style=None, childs=None, **initial_context):
        self.attrs = self.build_attrs(attrs)
        self.classes = self.build_classes(classes)
        self.childs = self.build_childs(childs)
        self.style = self.build_style(style)
        self.initial_context = self.build_context_data(initial_context)

    def __str__(self):
        return str(self.__class__)

    def __repr__(self):
        return "<%(cls)s childs=(%(childs)s)>" % {
            "cls": self.__class__.__name__,
            "childs": ";".join(self.childs),
        }

    def __getitem__(self, index):
        try:
            return self.childs[index]
        except KeyError:
            raise KeyError(
                "Key '%s' not found in '%s'. Choices are: %s."
                % (
                    index,
                    self.__class__.__name__,
                    ", ".join(sorted(self.childs)),
                )
            )

    def build_classes(self, classes):
        if classes is None:
            classes = []
        elif isinstance(classes, list):
            classes = classes.copy()
        elif isinstance(classes, str):
            classes = classes.split(" ")
        else:
            raise ValueError("%s is not valid valid classes." % classes)
        return classes

    def get_classes(self, extra_classes=None):
        new_classes = self.build_classes(extra_classes)
        return " ".join(self.classes + new_classes)

    def build_attrs(self, attrs):
        if attrs is None:
            attrs = {}
        elif not isinstance(attrs, dict):
            raise ValueError("%s is not valid valid attrs." % type(attrs))
        else:
            pass
        return attrs

    def get_attrs(self, extra_attrs=None):
        """Build an attribute dictionary."""
        extra = self.build_attrs(extra_attrs)
        attrs = {**self.attrs, **extra}
        return flatatt(attrs)

    def build_style(self, style):
        if style is None:
            style = {}
        elif not isinstance(style, dict):
            raise ValueError("%s is not valid valid style." % type(style))
        else:
            pass
        return style

    def get_style(self, extra_style=None):
        """Build an attribute dictionary."""
        extra = self.build_attrs(extra_style)
        attrs = {**self.style, **extra}
        return flatstyle(attrs)

    def build_childs(self, childs):
        if childs is None:
            _childs = []
        elif isinstance(childs, BaseComponent):
            _childs = [childs]
        elif isinstance(childs, (tuple, list)):
            _childs = list(childs)
        else:
            raise ValueError("%s is not valid childs argument type." % type(childs))
        for child in _childs:
            assert isinstance(child, BaseComponent), (
                "childs item %s must be an instance of Component class " % child
            )
        return _childs

    def get_childs(self, parent_context):
        rendered_childs = ""
        for child in self.childs:
            rendered_childs += child.render(context=parent_context)
        return mark_safe(rendered_childs)

    def get_template(self, template_name=None):
        """Return template name"""
        if not (template_name or self.template_string or self.template_name):
            class_name = self.__class__.__name__
            raise InvalidComponentTemplate(
                "template_string or template_name is not set for atomic component %s" % class_name
            )
        if self.template_string:
            template = Template(template_string=self.get_template_string())
            return template
        if not template_name:
            return get_template(template_name or self.get_template_name())

    def build_context_data(self, context):
        if context is None:
            context = {}
        elif not isinstance(context, dict):
            raise ValueError("%s is not valid valid context dict." % context)
        else:
            pass
        return context

    def get_context_data(self, **kwargs):
        return kwargs

    def get_template_string(self):
        return str(self.template_string)

    def get_template_name(self, context=None):
        return construct_template_name(self.template_name)

    def is_shown(self, context=None):
        return True

    def render(self, context=None, extra_attrs=None, extra_style=None, extra_classes=None):
        if self.is_shown(context):
            # set default component context
            context_base = {
                "attrs": self.get_attrs(extra_attrs=None),
                "classes": self.get_classes(extra_classes=None),
                "style": self.get_style(extra_style=None),
                "childs": self.get_childs(context),
                **self.initial_context,
            }
            # update base context based with render context
            context_data = self.get_context_data(**context_base)
            if context_data is None:
                raise TypeError("Expected a dict from get_context_data, got None")
            render_method = getattr(self, "render_html")
            return render_method(context_data)
        else:
            return ""

    def render_html(self, context):
        template = self.get_template()
        return template.render(Context(context))

    def render_json(self, context):
        raise NotImplementedError("%s doesn't implement render_json method" % self.__class__.__name__)

    def render_dict(self, context):
        raise NotImplementedError("%s doesn't implement render_dict method" % self.__class__.__name__)


class ComponentMetaclass(MediaDefiningClass):
    """Collect child components declared on the base classes."""

    def __new__(mcs, name, bases, attrs):
        # Collect childs from current class and remove them from attrs.
        attrs["declared_childs"] = {
            key: attrs.pop(key) for key, value in list(attrs.items()) if isinstance(value, BaseComponent)
        }

        new_class = super().__new__(mcs, name, bases, attrs)

        # Walk through the MRO.
        declared_childs = {}
        for base in reversed(new_class.__mro__):
            # Collect fields from base class.
            if hasattr(base, "declared_childs"):
                declared_childs.update(base.declared_childs)

            # Field shadowing.
            for attr, value in base.__dict__.items():
                if value is None and attr in declared_childs:
                    declared_childs.pop(attr)

        new_class.base_childs = declared_childs
        new_class.declared_childs = declared_childs

        return new_class


class Component(BaseComponent):
    pass



from django_atomics.component import Component


class Container(Component):
    template_string = """
    <div {% if attrs %}{{ attrs }}{% endif %} class="container{% if classes %} {{ classes }}{% endif %}">
        {{ childs }}
    </div>
    """


class Row(Component):
    template_string = """
    <div {% if attrs %}{{ attrs }}{% endif %} class="row{% if classes %} {{ classes }}{% endif %}">
        {{ childs }}
    </div>
    """


class Col(Component):
    template_string = """
    <div {% if attrs %}{{ attrs }}{% endif %} class="col{% if classes %} {{ classes }}{% endif %}">
        {{ childs }}
    </div>
    """



from django_atomics.component import Component


class Div(Component):
    template_string = """
    <div {% if attrs %}{{ attrs }}{% endif %}
         class="{% if classes %}{{ classes }}{% endif %}">{{ childs }}</div>
    """


class Nav(Component):
    template_string = """
    <nav {% if attrs %}{{ attrs }}{% endif %}
         class="{% if classes %}{{ classes }}{% endif %}">{{ childs }}</nav>
    """


class Header(Component):
    template_string = """
    <header {% if attrs %}{{ attrs }}{% endif %}
         class="{% if classes %}{{ classes }}{% endif %}">{{ childs }}</header>
    """


class Footer(Component):
    template_string = """
    <footer {% if attrs %}{{ attrs }}{% endif %}
         class="{% if classes %}{{ classes }}{% endif %}">{{ childs }}</footer>
    """


class Article(Component):
    template_string = """
    <article {% if attrs %}{{ attrs }}{% endif %}
         class="{% if classes %}{{ classes }}{% endif %}">{{ childs }}</article>
    """


class Section(Component):
    template_string = """
    <section {% if attrs %}{{ attrs }}{% endif %} class="section{% if classes %} {{ classes }}{% endif %}">
        {{ childs }}
    </section>
    """


class UnOrderedList(Component):
    template_string = """
    <ul {% if attrs %}{{ attrs }}{% endif %}
         class="{% if classes %}{{ classes }}{% endif %}">{{ childs }}</ul>
    """


class OrderedList(Component):
    template_string = """
    <ol {% if attrs %}{{ attrs }}{% endif %}
         class="{% if classes %}{{ classes }}{% endif %}">{{ childs }}</ol>
    """


class ListItem(Component):
    template_string = """
    <li {% if attrs %}{{ attrs }}{% endif %}
         style="{{ style }}",
         class="{% if classes %}{{ classes }}{% endif %}">{{ childs }}</li>
    """



from django_atomics.component import Component


class Content(Component):
    template_string = """{{ value|safe }}"""


class Paragraph(Component):
    template_string = """<p>{{ text|safe }}</p>"""


class Icon(Component):
    template_string = """<i {{ attrs }} class="{{ classes }}" href="{{ url }}"></i> """


class MaterialIcon(Icon):
    template_string = """<i {{ attrs }} class="mdi mdi-{{ name }} {{ classes}}" href="{{ url }}"></i> """


class BoostrapIcon(Icon):
    template_string = """<i {{ attrs }} class="bi bi-{{ name }} {{ classes}}" href="{{ url }}"></i> """


class FontAwesomeIcon(Icon):
    template_string = """<i {{ attrs }} class="fa fa-{{ name }} {{ classes}}" href="{{ url }}"></i> """


class Link(Component):
    template_string = """<a {{ attrs }} class="{{ classes}}" href="{{ url }}">{{ label }}</a> """


class Button(Component):
    template_string = """
    <button {{ attrs }} class="{{ classes}}"
            type="{{ type }}" {% if action %}formaction="{{ action }}"{% endif %}>
            {{ label }}
    </button>
    """


class ButtonLink(Component):
    template_string = """
    <a href="{{ url }}" class="btn btn{{ variant }}" target="{{ target }}" >{{ text }}</a>
    """



from django.utils.html import format_html_join


def flatatt(attrs):
    """
    Convert a dictionary of attributes to a single string.
    The returned string will contain a leading space followed by key="value",
    XML-style pairs. In the case of a boolean value, the key will appear
    without a value. It is assumed that the keys do not need to be
    XML-escaped. If the passed dictionary is empty, then return an empty
    string.

    The result is passed through 'mark_safe' (by way of 'format_html_join').
    """
    key_value_attrs = []
    boolean_attrs = []
    for attr, value in attrs.items():
        if isinstance(value, bool):
            if value:
                boolean_attrs.append((attr,))
        elif value is not None:
            key_value_attrs.append((attr, value))

    return format_html_join("", ' {}="{}"', sorted(key_value_attrs)) + format_html_join(
        "", " {}", sorted(boolean_attrs)
    )


def flatstyle(style):
    """
    Convert a dictionary of attributes to a single string.
    The returned string will contain property:value; CSS Style.
    The result is passed through 'mark_safe' (by way of 'format_html_join').
    """

    prop_values = []
    for property, prop_value in style.items():
        prop_values.append((property, prop_value))
    return format_html_join("", '{}:{}; ', sorted(prop_values))


def construct_template_name(template_name):
    return template_name



from django.template import Library
from django_atomics.component import Component, NotComponent

register = Library()


@register.simple_tag(name="component", takes_context=True)
def render_component(context, component, **extra_context):
    if not isinstance(component, Component):
        raise NotComponent("%s is not component!")
    context = {"request": context["request"], **extra_context}
    return component.render(context)



def is_admin(user):
    return user.is_superuser or user.is_staff


class RequestContextEmpty(Exception):
    pass


class Permission:
    def has_permission(self, user):
        return True

    def has_object_permission(self, user, obj):
        return True

    def check_permission(self, user, obj=None):
        perm = self.has_permission(user)
        if not perm:
            return False
        # Check object permission
        if obj:
            obj_perm = self.has_object_permission(user, obj)
            if not obj_perm:
                perm = obj_perm
        return perm


class IsAuthenticated(Permission):
    def has_permission(self, user):
        return user.is_authenticated


class IsAdmin(Permission):
    def has_permission(self, user):
        return is_admin(user)


class IsSuperadmin(Permission):
    def has_permission(self, user):
        return user.is_superuser


class PermissionMixin:

    permission_classes = []
    model_instance_name = "instance"

    def get_permission_classes(self):
        return self.permission_classes

    def get_permissions(self):
        permissions = [permission() for permission in self.get_permission_classes()]
        return permissions

    def has_permissions(self, user, obj=None):
        has_perm = True
        permissions = self.get_permissions()
        for permission in permissions:
            perm = permission.check_permission(user, obj)
            if not perm:
                has_perm = False
        return has_perm

    def is_shown(self, context=None):
        if context is None:
            raise RequestContextEmpty("Component permission need request context!")
        request = context["request"]
        instance = context.get(self.model_instance_name, None)
        user = request.user
        return self.has_permissions(user, instance)


class AuthenticatedMixin(PermissionMixin):

    permission_classes = [IsAuthenticated]


class AdminOnlyMixin(PermissionMixin):

    permission_classes = [IsAdmin]



from django_atomics.component import Component


class Img(Component):
    template_string = """
    <img {{ attrs }}
         src="{{ url }}"
         {% if width %} width={{ width }}{% endif %}
         {% if height %} height={{ height }}{% endif %}
         class="{{ classes }}"
         alt="{{ alt }}"/>
    """
